#!/usr/bin/env python
# -*- coding: utf-8 -*-

import asyncio
from .manager import manager


class Motor:
    @classmethod
    async def create(cls, address, name, update=None, *, loop=None):
        motor = cls()
        await motor._connect(address, name, update, loop=loop)
        return motor

    async def _connect(self, address, name, update, *, loop=None):
        self._loop = loop or asyncio.get_event_loop()
        self._address = address
        self._name = name
        self._manager = manager
        self._update = update
        self._position = 0
        self._connection = await self._manager.connect(self._address, loop=self._loop)
        self._message = self._connection.message
        await self._connection.send(self._message.motor_register_position(self._name))
        self.update()

    def update(self):
        if callable(self._update):
            self._update(self._name, self._position)

    async def get_position(self):
        _, self._position = await self._connection.send(self._message.motor_read_position(self._name))
        self.update()
        return self._position

    def __del__(self):
        self._connection.close()
        self._connection = None
        self._message = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import struct


MAGIC_NUMBER = 4277009102
NATIVE_HEADER_VERSION = 4
HEADER_FORMAT = '<IiIIIIiiIIIii80s'
HEADER_SIZE = struct.calcsize(HEADER_FORMAT)
CLOSE = 1
ABORT = 2
CMD = 3
CMD_WITH_RETURN = 4
REGISTER = 6
UNREGISTER = 7
EVENT = 8
FUNC = 9
FUNC_WITH_RETURN = 10
CHAN_READ = 11
CHAN_SEND = 12
REPLY = 13
HELLO = 14
HELLO_REPLY = 15
TYPE_DOUBLE = 1
TYPE_STRING = 2
TYPE_ERROR = 3
TYPE_ASSOC = 4
MAX_CHUNK_SIZE = 65536
SPEC_MIN_PORT = 6510
SPEC_MAX_PORT = 6530



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from . import const
from .header import LHeader


class Message:
    def __init__(self):
        self._serial = -1

    @property
    def serial(self):
        return self._serial

    @serial.setter
    def serial(self, serial):
        self._serial = serial

    def _pack_header(self, cmd, datalen=0, name='', *, typ=const.TYPE_STRING):
        self._serial += 1
        h = LHeader()
        packed = h.pack(self._serial, cmd, typ, datalen, name)
        return packed

    def _pack_data(self, data):
        return b'%s\x00' % str(data).encode(encoding='ascii', errors='ignore')

    def hello(self):
        return self._pack_header(const.HELLO, name='python')

    def abort(self):
        return self._pack_header(const.ABORT)

    def motor_register_position(self, name):
        s = 'motor/{}/position'.format(name)
        return self._pack_header(const.REGISTER, name=s)

    def motor_register_high_limit_hit(self, name):
        s = 'motor/{}/high_limit_hit'.format(name)
        return self._pack_header(const.REGISTER, name=s)

    def motor_register_low_limit_hit(self, name):
        s = 'motor/{}/low_limit_hit'.format(name)
        return self._pack_header(const.REGISTER, name=s)

    def motor_register_high_limit(self, name):
        s = 'motor/{}/high_limit'.format(name)
        return self._pack_header(const.REGISTER, name=s)

    def motor_register_low_limit(self, name):
        s = 'motor/{}/low_limit'.format(name)
        return self._pack_header(const.REGISTER, name=s)

    def motor_register_move_done(self, name):
        s = 'motor/{}/move_done'.format(name)
        return self._pack_header(const.REGISTER, name=s)

    def motor_read_position(self, name):
        s = 'motor/{}/position'.format(name)
        return self._pack_header(const.CHAN_READ, name=s)

    def motor_move(self, name, position):
        data = self._pack_data(position)
        s = 'motor/{}/start_one'.format(name)
        header = self._pack_header(const.CHAN_SEND, len(data), s)
        return header + data

    def counter_register(self, name):
        s = 'scaler/{}/value'.format(name)
        return self._pack_header(const.REGISTER, name=s)

    def counter_read(self, name):
        s = 'scaler/{}/value'.format(name)
        return self._pack_header(const.CHAN_READ, name=s)

    def counter_count(self, name, sec):
        data = self._pack_data(sec)
        s = 'scaler/{}/count'.format(name)
        header = self._pack_header(const.CHAN_SEND, len(data), s)
        return header + data

    def command_run(self, command):
        data = self._pack_data(command)
        header = self._pack_header(const.CMD_WITH_RETURN, len(data))
        return header + data

    def command_register(self):
        s = 'status/ready'
        header = self._pack_header(const.REGISTER, name=s)
        return header



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from PyQt5 import QtCore
from . import const


class Qommand(QtCore.QObject):
    sigFinished = QtCore.pyqtSignal(object)

    def __init__(self, connection):
        super().__init__()
        self._serial = -1
        self._connection = connection
        self._message = self._connection.message()
        self._connection.sigSpecReplyArrived.connect(self._parseReply)
        self._connection.sigConnectedToSpec.connect(self.register)

    def run(self, command):
        self._connection.send(self._message.command_run(command))
        self._serial = self._message.serial

    def register(self):
        self._connection.send(self._message.command_register())

    def _parseReply(self, header, value):
        if header.typ == const.TYPE_ERROR:
            self._connection.sigError.emit(value)
            return
        if self._serial == header.id:
            self._serial = -1
            self.sigFinished.emit(value)
            return



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import asyncio
from collections import deque
from . import const, excepts
from .reply import Reply
from .message import Message


async def _connect_to_spec(address, *, loop=None):
    host, port = address
    message = Message()
    reader, writer = await asyncio.open_connection(host, port, loop=loop)
    conn = SpecConnection(reader, writer, message, address=address, loop=loop)
    await conn.get_name()
    return conn


async def _search_for_spec(address, *, loop=None):
    host, spec = address
    for port in range(const.SPEC_MIN_PORT, const.SPEC_MAX_PORT):
        conn = await _connect_to_spec((host, port), loop=loop)
        if conn.name == spec:
            return conn
        else:
            conn.close()
    raise excepts.SpecConnectionError('Could not find such spec connection')


async def create_connection(address, *, loop=None):
    host, spec = address
    loop = loop or asyncio.get_event_loop()
    try:
        int(spec)
    except ValueError:
        return await _search_for_spec(address, loop=loop)
    return await _connect_to_spec(address, loop=loop)


class SpecConnection:
    def __init__(self, reader, writer, message, *, address, loop=None):
        self.name = ''
        self._reader = reader
        self._writer = writer
        self.message = message
        self._address = address
        self._loop = loop
        self._waiters = deque()
        self._reply = Reply()
        self._reader_task = asyncio.ensure_future(self._read_data(), loop=self._loop)

    async def get_name(self):
        header, name = await self.send(self.message.hello())
        self.name = name

    def __repr__(self):
        host, port = self._address
        return '<Async SpecConnection to {} running at {}:{}>'.format(self.name, host, port)

    async def _read_data(self):
        while not self._reader.at_eof():
            try:
                data = await self._reader.read(const.MAX_CHUNK_SIZE)
            except asyncio.CancelledError:
                break
            self._process_answer(data)

    def _process_answer(self, data):
        for header, answer in self._reply.unpack(data):
            waiter, = self._waiters.popleft()
            waiter.set_result((header, answer))

    def send(self, message):
        if self._reader is None or self._reader.at_eof():
            raise excepts.SpecConnectionError('Spec connection was closed')
        fut = self._loop.create_future()
        self._writer.write(message)
        self._waiters.append((fut,))
        return fut

    def close(self):
        self._do_close()

    def _do_close(self):
        self._writer.transport.close()
        self._reader_task.cancel()
        self._reader_task = None
        self._writer = None
        self._reader = None
        self._waiters = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import asyncio
from .connection import create_connection
from .qonnection import Qonnection


class _Manager:
    def __init__(self):
        self._connections = {}

    async def connect(self, address, *, loop=None):
        self._loop = loop or asyncio.get_event_loop()
        if address not in self._connections:
            self._connections[address] = await create_connection(address, loop=self._loop)
        return self._connections[address]

    def qonnect(self, address):
        if address not in self._connections:
            self._connections[address] = Qonnection(address)
        return self._connections[address]

    def disconnect(self, address):
        if address in self._connections:
            self._connections[address].close()
            del self._connections[address]

    def abort(self):
        for address in self._connections:
            self._connections[address].abort()


manager = _Manager()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import time
import ctypes
from . import const


class _Header:
    _fields_ = [
        ('magic', ctypes.c_uint32),
        ('header', ctypes.c_int32),
        ('size', ctypes.c_uint32),
        ('id', ctypes.c_uint32),
        ('sec', ctypes.c_uint32),
        ('usec', ctypes.c_uint32),
        ('cmd', ctypes.c_int32),
        ('typ', ctypes.c_int32),
        ('rows', ctypes.c_uint32),
        ('cols', ctypes.c_uint32),
        ('datalen', ctypes.c_uint32),
        ('error', ctypes.c_int32),
        ('flags', ctypes.c_int32),
        ('name', ctypes.c_char * 80)
    ]

    def pack(self, serial, cmd, typ, datalen, name):
        timestamp = time.time()
        self.magic = const.MAGIC_NUMBER
        self.header = const.NATIVE_HEADER_VERSION
        self.size = const.HEADER_SIZE
        self.id = serial
        self.sec = int(timestamp)
        self.usec = int((timestamp - self.sec) * 1e6)
        self.cmd = cmd
        self.typ = typ
        self.datalen = datalen
        self.name = name.encode()
        self.error = 0
        return bytes(self)

    def unpack(self, buffer):
        ctypes.memmove(ctypes.addressof(self), buffer, const.HEADER_SIZE)
        return self

    def __repr__(self):
        try:
            return ('<Spec header: size={:d}, id={:d}, cmd={:d}, type={:d}, datalen={:d}, error={:d}, name={}'
                    ''.format(self.size, self.id, self.cmd, self.typ, self.datalen, self.error, self.name))
        except AttributeError:
            return ''


class LHeader(ctypes.LittleEndianStructure):
    _fields_ = _Header._fields_
    pack = _Header.pack
    unpack = _Header.unpack
    __repr__ = _Header.__repr__


class BHeader(ctypes.BigEndianStructure):
    _fields_ = _Header._fields_
    pack = _Header.pack
    unpack = _Header.unpack
    __repr__ = _Header.__repr__


Header = LHeader



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from PyQt5 import QtCore
from . import const
from .manager import manager


class Qounter(QtCore.QObject):
    sigConnected = QtCore.pyqtSignal(str)
    sigValueChanged = QtCore.pyqtSignal(str, float)
    sigError = QtCore.pyqtSignal(str)

    def __init__(self, address, name):
        super().__init__()
        self._address = address
        self._name = name
        self._value = 0
        self._manager = manager
        self._connected = False
        self._connect()

    def _connect(self):
        self._connection = self._manager.qonnect(self._address)
        self._message = self._connection.message()
        self._connection.sigConnectedToSpec.connect(self._connectedToSpec)
        self._connection.sigSpecReplyArrived.connect(self._parseReply)
        self._connection.sigError.connect(self._connectionHasError)
        if self._connection.isConnected():
            self._connectedToSpec()

    def _connectionHasError(self, emsg):
        self._connected = False
        self.sigError.emit(emsg)

    def _connectedToSpec(self):
        self._connected = True
        self._connection.send(self._message.counter_register(self._name))
        self._connection.send(self._message.counter_read(self._name))
        self.sigConnected.emit(self._name)

    def _parseReply(self, header, value):
        try:
            device, name, propert = header.name.decode().split('/')
        except ValueError:
            return
        if device != 'scaler' or name != self._name or header.cmd not in (const.EVENT, const.REPLY):
            return
        if propert == 'value':
            self._value = value
            self.sigValueChanged.emit(self._name, value)

    def read(self):
        self._connection.send(self._message.counter_read(self._name))

    def count(self, sec):
        self._connection.send(self._message.counter_count(self._name, sec))

    def isConnected(self):
        return self._connected

    def name(self):
        return self._name

    def value(self):
        return self._value

    def connection(self):
        return self._connection



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from PyQt5 import QtCore, QtNetwork
from . import const
from .reply import Reply
from .message import Message
from .qommand import Qommand


class Qonnection(QtCore.QObject):
    sigConnectedToSpec = QtCore.pyqtSignal()
    sigSpecReplyArrived = QtCore.pyqtSignal(object, object)
    sigError = QtCore.pyqtSignal(str)

    def __init__(self, address, reconnect=True):
        super().__init__()
        self._address = address
        self._connected = False
        self._reconnect = reconnect
        self._connect()

    def _connect(self):
        host, port = self._address
        self._host = host
        try:
            self._port = int(port)
        except ValueError:
            self._searching = True
            self._port = const.SPEC_MIN_PORT
            self._name = port
        else:
            self._name = ''
            self._searching = False
        self._connectToSpec()

    def _connectToSpec(self):
        self._message = Message()
        self._reply = Reply()
        self._command = Qommand(self)
        self._setSocket()
        self._connectSignals()
        self._sock.connectToHost(self._host, self._port)

    def send(self, message):
        s = self._sock.state()
        if s == QtNetwork.QAbstractSocket.ConnectedState:
            self._sock.write(message)
        elif s == QtNetwork.QAbstractSocket.UnconnectedState and not self._connected and not self._searching:
            if self._reconnect:
                self._connectToSpec()
            else:
                self.sigError.emit('Lost connection to {} at {}:{}'.format(self._name, self._host, self._port))
        else:
            self.sigError.emit('Trying to send messages without established connection to {}'.format(self._name))

    def _setSocket(self):
        self._sock = QtNetwork.QTcpSocket(self)
        self._sock.setProxy(QtNetwork.QNetworkProxy(QtNetwork.QNetworkProxy.NoProxy))

    def _connectSignals(self):
        self._sock.connected.connect(self._sayHello)
        self._sock.readyRead.connect(self._readResponse)
        self._sock.disconnected.connect(self._stopConnection)
        self._sock.error.connect(self._serverHasError)

    def _sayHello(self):
        self.send(self._message.hello())

    def _readResponse(self):
        buf = bytes(self._sock.readAll())
        if self._searching and not self._connected:
            for header, answer in self._reply.unpack(buf):
                if header.cmd == const.HELLO_REPLY:
                    if answer == self._name:
                        self._connected = True
                        self._searching = False
                        self._address = self._host, self._port
                        self.sigConnectedToSpec.emit()
                    else:
                        self._port += 1
                        if self._port >= const.SPEC_MAX_PORT:
                            self._connected = False
                            self._searching = False
                            self.sigError.emit('Count not find {} on the host {}'.format(self._name, self._host))
                            return
                        self._connectToSpec()
        else:
            for header, answer in self._reply.unpack(buf):
                if header.cmd == const.HELLO_REPLY:
                    self._connected = True
                    self._searching = False
                    self._name = answer
                    self.sigConnectedToSpec.emit()
                else:
                    self.sigSpecReplyArrived.emit(header, answer)

    def _stopConnection(self):
        self._searching = False
        self._connected = False
        self.sigError.emit('The connection to {} at {}:{} has been stopped'.format(self._name, self._host, self._port))

    def _serverHasError(self):
        self._connected = False
        self._searching = False
        e = self._sock.errorString()
        self.sigError.emit('The server {} at {}:{} has an error: {}'.format(self._name, self._host, self._port, e))

    def close(self):
        self._stopConnection()
        self._sock.close()

    def __repr__(self):
        return '<Qt SpecConnection to {} running at {}:{}>'.format(self._name, self._host, self._port)

    def isConnected(self):
        return (not self._searching and self._connected and
                QtNetwork.QAbstractSocket.ConnectedState == self._sock.state())

    def name(self):
        return self._name

    def message(self):
        return self._message

    def abort(self):
        self.send(self._message.abort())

    def command(self):
        return self._command



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

setup(
    name='aspic',
    version='2017.3.16.1',
    description='Asyncio and Qt versions of SpecClient for Python 3',
    author='Vadim Dyadkin',
    author_email='dyadkin@gmail.com',
    url='https://hg.3lp.cx/aspic',
    license='GPLv3',
    package_dir={'aspic': ''},
    py_modules=[
        'aspic.__init__',
        'aspic.connection',
        'aspic.const',
        'aspic.excepts',
        'aspic.header',
        'aspic.manager',
        'aspic.message',
        'aspic.motor',
        'aspic.qommand',
        'aspic.qmotor',
        'aspic.qonnection',
        'aspic.qounter',
        'aspic.reply',
    ],
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from . import const
from . header import LHeader, BHeader


class Reply:
    def __init__(self):
        self._serial = -1
        self._buffer = b''
        self._header = None

    def unpack(self, buf):
        self._buffer += buf
        while self._buffer:
            self._unpack_header()
            data = self._unpack_body()
            if data is None:
                raise StopIteration
            data = self._convert_data(data)
            header = self._header
            self._header = None
            yield header, data

    def _unpack_body(self):
        if not self._header or len(self._buffer) < self._header.datalen:
            return
        data = self._buffer[:self._header.datalen-1]  # cut b'\x00' here
        self._buffer = self._buffer[self._header.datalen:]
        return data

    def _unpack_header(self):
        while not self._header and len(self._buffer) >= const.HEADER_SIZE:
            header = LHeader()
            header.unpack(self._buffer[:const.HEADER_SIZE])
            if header.magic != const.MAGIC_NUMBER:
                header = BHeader()
                header.unpack(self._buffer[:const.HEADER_SIZE])
                if header.magic != const.MAGIC_NUMBER:
                    self._buffer = self._buffer[1:]
                    continue
            header.name = header.name.replace(b'\x00', b'')
            self._buffer = self._buffer[const.HEADER_SIZE:]
            self._header = header

    def _convert_data(self, data):
        if self._header.typ in (const.TYPE_STRING, const.TYPE_DOUBLE):
            try:
                data = float(data)
            except ValueError:
                data = data.decode()
            return data
        else:
            return data.decode()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from PyQt5 import QtCore
from .manager import manager
from . import const


class Motor(QtCore.QObject):
    sigConnected = QtCore.pyqtSignal(str)
    sigNewPosition = QtCore.pyqtSignal(str, float)
    sigMoveDone = QtCore.pyqtSignal(str)
    sigLimitHit = QtCore.pyqtSignal(str)
    sigError = QtCore.pyqtSignal(str)

    def __init__(self, address, name):
        super().__init__()
        self._manager = manager
        self._address = address
        self._name = name
        self._position = 0
        self._low_limit = 0
        self._high_limit = 0
        self._on_limit = False
        self._connected = False
        self._moving = False
        self._connect()

    def __repr__(self):
        return 'Qt Motor {} with {}'.format(self._name, self._connection)

    def _connect(self):
        self._connection = self._manager.qonnect(self._address)
        self._message = self._connection.message()
        self._connection.sigConnectedToSpec.connect(self._connectedToSpec)
        self._connection.sigSpecReplyArrived.connect(self._parseReply)
        self._connection.sigError.connect(self._connectionHasError)
        if self._connection.isConnected():
            self._connectedToSpec()

    def _connectionHasError(self, emsg):
        self._connected = False
        self.sigError.emit(emsg)

    def _connectedToSpec(self):
        self._connected = True
        self._connection.send(self._message.motor_register_high_limit_hit(self._name))
        self._connection.send(self._message.motor_register_low_limit_hit(self._name))
        self._connection.send(self._message.motor_register_high_limit(self._name))
        self._connection.send(self._message.motor_register_low_limit(self._name))
        self._connection.send(self._message.motor_register_position(self._name))
        self._connection.send(self._message.motor_register_move_done(self._name))
        self._connection.send(self._message.motor_read_position(self._name))
        self.sigConnected.emit(self._name)

    def _parseReply(self, header, value):
        try:
            device, name, propert = header.name.decode().split('/')
        except ValueError:
            return
        if device != 'motor' or name != self._name or header.cmd not in (const.EVENT, const.REPLY):
            return
        if propert == 'position':
            self._position = value
            self._on_limit = self._position in (self._low_limit, self._high_limit)
            self.sigNewPosition.emit(self._name, self._position)
        elif propert == 'move_done' and value == 0:
            self._on_limit = self._position in (self._low_limit, self._high_limit)
            self._moving = False
            self.sigMoveDone.emit(self._name)
        elif propert == 'move_done' and value != 0:
            self._moving = True
        elif propert in ('high_limit_hit', 'low_limit_hit'):
            self._on_limit = True
        elif propert == 'low_limit':
            self._low_limit = value
        elif propert == 'high_limit':
            self._high_limit = value
        if self._on_limit:
            self.sigLimitHit.emit(self._name)

    def move(self, position):
        if position == self._position:
            return
        self._connection.send(self._message.motor_move(self._name, position))

    def moveRelative(self, position):
        self.move(self._position + position)

    def stop(self):
        self._connection.abort()

    def name(self):
        return self._name

    def position(self):
        return self._position

    def isOnLimit(self):
        return self._on_limit

    def highLimit(self):
        return self._high_limit

    def lowLimit(self):
        return self._low_limit

    def isConnected(self):
        return self._connected

    def connection(self):
        return self._connection

    def manager(self):
        return self._manager

    def isMoving(self):
        return self._moving
